import logging

from twisted.internet import reactor

from peek_core_device._private.storage.DeviceInfoTuple import DeviceInfoTuple
from peek_core_device._private.storage.DeviceUpdateTuple import DeviceUpdateTuple
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

logger = logging.getLogger(__name__)


class NotifierController:

    def __init__(self, tupleObservable: TupleDataObservableHandler):
        self._tupleObservable = tupleObservable

        from peek_core_device._private.server.DeviceApi import DeviceApi
        self._api: DeviceApi = None

    def setApi(self, api):
        self._api = api

    def shutdown(self):
        self._tupleObservable = None
        self._api = None

    def notifyDeviceInfo(self, deviceId: str):
        reactor.callLater(0, self._notifyDeviceInfoObservable, deviceId)

    def _notifyDeviceInfoObservable(self, deviceId: str):
        """ Notify the observer of the update

         This tuple selector must exactly match what the UI observes

        """

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceInfoTuple.tupleName(), dict(deviceId=deviceId))
        )

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceInfoTuple.tupleName(), dict())
        )

    def notifyDeviceUpdate(self, deviceType: str):
        reactor.callLater(0, self._notifyDeviceUpdateObservable, deviceType)

    def _notifyDeviceUpdateObservable(self, deviceType: str):
        """ Notify the observer of the update

         This tuple selector must exactly match what the UI observes

        """

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceUpdateTuple.tupleName(), dict(deviceType=deviceType))
        )

        self._tupleObservable.notifyOfTupleUpdate(
            TupleSelector(DeviceUpdateTuple.tupleName(), dict())
        )

    def notifyDeviceOnline(self, deviceId: str, deviceToken: str, online: bool):
        """ Notify Device Online

        Notify that the device has changed it's online status

        """
        reactor.callLater(0, self._notifyDeviceOnlineObservable,
                          deviceId, deviceToken, online)

    def _notifyDeviceOnlineObservable(self, deviceId: str, deviceToken: str,
                                      online: bool):
        self._api.notifyOfOnlineStatus(deviceId, deviceToken, online)



from peek_core_device._private.PluginNames import deviceFilt
from peek_core_device._private.PluginNames import deviceObservableName
from peek_core_device._private.server.tuple_providers.ClientSettingsTupleProvider import \
    ClientSettingsTupleProvider
from peek_core_device._private.server.tuple_providers.DeviceInfoTupleProvider import \
    DeviceInfoTupleProvider
from peek_core_device._private.server.tuple_providers.DeviceUpdateTupleProvider import \
    DeviceUpdateTupleProvider
from peek_core_device._private.storage.DeviceInfoTuple import DeviceInfoTuple
from peek_core_device._private.storage.DeviceUpdateTuple import DeviceUpdateTuple
from peek_core_device._private.tuples.ClientSettingsTuple import ClientSettingsTuple
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler


def makeTupleDataObservableHandler(ormSessionCreator):
    """" Make Tuple Data Observable Handler

    This method creates the observable object, registers the tuple providers and then
    returns it.

    :param ormSessionCreator: A function that returns a SQLAlchemy session when called

    :return: An instance of :code:`TupleDataObservableHandler`

    """
    tupleObservable = TupleDataObservableHandler(
        observableName=deviceObservableName,
        additionalFilt=deviceFilt)

    # Register TupleProviders here
    tupleObservable.addTupleProvider(DeviceUpdateTuple.tupleName(),
                                     DeviceUpdateTupleProvider(ormSessionCreator))

    tupleObservable.addTupleProvider(DeviceInfoTuple.tupleName(),
                                     DeviceInfoTupleProvider(ormSessionCreator))

    tupleObservable.addTupleProvider(ClientSettingsTuple.tupleName(),
                                     ClientSettingsTupleProvider(ormSessionCreator))

    return tupleObservable



from peek_core_device._private.PluginNames import deviceTuplePrefix
from vortex.Tuple import addTupleType, TupleField, Tuple


@addTupleType
class ClientSettingsTuple(Tuple):
    """ Client Settings Tuple

    This tuple is for the client UI settings.

    """
    __tupleType__ = deviceTuplePrefix + "ClientSettingsTuple"

    fieldEnrollmentEnabled: bool = TupleField(defaultValue=False)
    officeEnrollmentEnabled: bool = TupleField(defaultValue=False)




# import shutil
# from pathlib import Path
#
# import logging
#
# from twisted.internet import defer
# from twisted.internet.defer import inlineCallbacks
#
# from peek_core_device._private.PluginNames import deviceFilt
# from txhttputil.downloader.HttpFileDownloader import HttpFileDownloader
# from vortex.DeferUtil import deferToThreadWrapWithLogger
#
# logger = logging.getLogger(__name__)
#
# _filt = dict(key="UpdateDownloadHandler")
# _filt.update(deviceFilt)
#
# class UpdateDownloadHandler(ModelHandler):
#     CHUNK_SIZE = 128000
#
#     def __init__(self, filtStoragePath:Path, serverHost:str, serverPort:int):
#         ModelHandler.__init__(self, payloadFilter=_filt)
#         self._filtStoragePath =  filtStoragePath
#         self._serverHost =  serverHost
#         self._serverPort =  serverPort
#
#     @inlineCallbacks
#     def buildModel(self, payloadFilt=None, vortexUuid=None, **kwargs):
#         chunk = payloadFilt["chunk"]
#         version = payloadFilt["version"]
#         deviceType = payloadFilt["deviceType"]
#         urlPath = payloadFilt["urlPath"]
#
#         localFile = self._filtStoragePath / deviceType / version
#
#         if localFile.is_file():
#             data = yield self._sendChunk(localFile, chunk)
#             defer.returnValue(data)
#
#         # If the file doesn't exist, download it from the server.
#         url = 'http://%(host)s:%(port)s/peek_core_device/device_update/%(urlPath)s'
#         url %= {'host':self._serverHost,
#                 'port':self._serverPort,
#                 'urlPath':urlPath}
#
#         logger.debug("Downloading file from %s" % url)
#         namedTempFile = yield HttpFileDownloader(url).run()
#         namedTempFile.delete = False
#         shutil.move(namedTempFile.name, str(localFile))
#
#         assert localFile.is_file(), "File download failed"
#
#         data = yield self._sendChunk(localFile, chunk)
#         defer.returnValue(data)
#
#     @deferToThreadWrapWithLogger(logger)
#     def _sendChunk(self, localFile:Path, chunk:int):
#         with localFile.open("rb") as f:
#             f.seek(chunk * self.CHUNK_SIZE)
#             return f.read(self.CHUNK_SIZE)


deviceFilt = {"plugin": "peek_core_device"}
deviceTuplePrefix = "peek_core_device."
deviceObservableName = "peek_core_device"
deviceActionProcessorName = "peek_core_device"
deviceTupleOfflineServiceName = "peek_core_device"


import logging
from typing import Optional, List

from rx import Observable
from rx.subjects import Subject
from twisted.internet.defer import Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger, noMainThread

from peek_core_device._private.server.controller.MainController import MainController
from peek_core_device._private.storage.DeviceInfoTuple import DeviceInfoTuple
from peek_core_device.server.DeviceApiABC import DeviceApiABC
from peek_core_device.tuples.DeviceDetailTuple import DeviceDetailTuple
from peek_core_device.tuples.DeviceOnlineDetailTuple import DeviceOnlineDetailTuple

logger = logging.getLogger(__name__)


class DeviceApi(DeviceApiABC):
    def __init__(self, mainController: MainController,
                 ormSessionCreator):
        self._mainController = mainController
        self._ormSessionCreator = ormSessionCreator

        self._deviceOnlineSubject = Subject()

    def shutdown(self):
        pass

    @deferToThreadWrapWithLogger(logger)
    def deviceDetails(self, deviceTokens: List[str]) -> Deferred:
        ormSession = self._ormSessionCreator()
        try:
            all = ormSession.query(DeviceInfoTuple) \
                .filter(DeviceInfoTuple.deviceToken.in_(deviceTokens)) \
                .all()

            tuples = [
                DeviceDetailTuple(
                    deviceToken=d.deviceToken,
                    deviceType=d.deviceType,
                    description=d.description,
                    lastOnline=d.lastOnline,
                    isOnline=d.isOnline,
                )
                for d in all
            ]

            return tuples

        finally:
            ormSession.close()

    @deferToThreadWrapWithLogger(logger)
    def deviceDescription(self, deviceToken: str) -> Deferred:
        return self.deviceDescriptionBlocking(deviceToken)

    def deviceDescriptionBlocking(self, deviceToken: str) -> Optional[str]:
        noMainThread()

        ormSession = self._ormSessionCreator()
        try:
            all = ormSession.query(DeviceInfoTuple) \
                .filter(DeviceInfoTuple.deviceToken == deviceToken) \
                .all()

            if not all:
                return None

            return all[0].description

        finally:
            ormSession.close()

    def deviceOnlineStatus(self) -> Observable:
        return self._deviceOnlineSubject

    def notifyOfOnlineStatus(self, deviceId: str, deviceToken: str, status: bool):
        self._deviceOnlineSubject.on_next(
            DeviceOnlineDetailTuple(
                deviceToken=deviceToken,
                deviceId=deviceId,
                onlineStatus=status
            )
        )



from peek_core_device._private.PluginNames import deviceTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class DeviceOnlineDetailTuple(Tuple):
    __tupleType__ = deviceTuplePrefix + "DeviceOnlineDetailTuple"

    deviceId: str = TupleField()
    deviceToken: str = TupleField()
    onlineStatus: bool = TupleField()



from peek_core_device._private.PluginNames import deviceTuplePrefix
from vortex.Tuple import addTupleType, Tuple, TupleField


@addTupleType
class DeviceDetailTuple(Tuple):
    __tupleType__ = deviceTuplePrefix + "DeviceDetailTuple"

    deviceToken: str = TupleField()
    deviceType: str = TupleField()
    description: str = TupleField()
    lastOnline: str = TupleField()
    isOnline: bool = TupleField()




import logging

from sqlalchemy import Column, Boolean
from sqlalchemy import Integer, String, DateTime
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import ForeignKey, Index

from .DeclarativeBase import DeclarativeBase
from vortex.Tuple import Tuple, addTupleType

from peek_core_device._private.PluginNames import deviceTuplePrefix

logger = logging.getLogger(__name__)


@addTupleType
class DeviceUpdateTuple(Tuple, DeclarativeBase):
    """ DeviceUpdateTuple

    This table stores information about the peek device updates.

    """
    __tablename__ = 'DeviceUpdate'
    __tupleType__ = deviceTuplePrefix + 'DeviceUpdateTuple'

    id = Column(Integer, primary_key=True)
    deviceType = Column(String(20), nullable=False)
    description = Column(String, nullable=False)
    buildDate = Column(DateTime(True), nullable=False)
    appVersion = Column(String(15), nullable=False)
    updateVersion = Column(String(15), nullable=False)
    filePath = Column(String(150), nullable=False)
    urlPath = Column(String(150), nullable=False)
    fileSize = Column(Integer, nullable=False)
    isEnabled = Column(Boolean, nullable=False, server_default="0")

    __table_args__ = (
        Index("idx_DeviceUpdate_Version",
              deviceType, appVersion, updateVersion, unique=True),
    )


import logging

from peek_core_device._private.PluginNames import deviceFilt
from peek_core_device._private.storage.Setting import SettingProperty, globalSetting
from peek_core_device._private.tuples.ClientSettingsTuple import ClientSettingsTuple
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler
from vortex.sqla_orm.OrmCrudHandler import OrmCrudHandler, OrmCrudHandlerExtension

logger = logging.getLogger(__name__)

# This dict matches the definition in the Admin angular app.
filtKey = {"key": "admin.Edit.SettingProperty"}
filtKey.update(deviceFilt)


# This is the CRUD handler
class __CrudHandler(OrmCrudHandler):
    # The UI only edits the global settings
    # You could get more complicated and have the UI edit different groups of settings.
    def createDeclarative(self, session, payloadFilt):
        return [p for p in globalSetting(session).propertyObjects]

class __ExtUpdateObservable(OrmCrudHandlerExtension):
    """ Update Observable ORM Crud Extension

    This extension is called after events that will alter data,
    it then notifies the observer.

    """
    def __init__(self, tupleDataObserver: TupleDataObservableHandler):
        self._tupleDataObserver = tupleDataObserver

    def _tellObserver(self, tuple_, tuples, session, payloadFilt):
        self._tupleDataObserver \
            .notifyOfTupleUpdate(TupleSelector(SettingProperty.tupleName(), {}))

        self._tupleDataObserver \
            .notifyOfTupleUpdate(TupleSelector(ClientSettingsTuple.tupleName(), {}))

        return True

    afterUpdateCommit = _tellObserver
    afterDeleteCommit = _tellObserver


# This method creates an instance of the handler class.
def makeSettingPropertyHandler(tupleObservable, dbSessionCreator):
    handler = __CrudHandler(dbSessionCreator, SettingProperty,
                            filtKey, retreiveAll=True)

    handler.addExtension(SettingProperty, __ExtUpdateObservable(tupleObservable))
    logger.debug("Started")
    return handler


"""timezone aware

Peek Plugin Database Migration Script

Revision ID: 6bf6ecef3108
Revises: 5dd43ff7ea8e
Create Date: 2017-12-31 14:28:47.653836

"""

# revision identifiers, used by Alembic.
revision = '6bf6ecef3108'
down_revision = '5dd43ff7ea8e'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import geoalchemy2


def upgrade():

    # core_device."DeviceInfo"
    op.execute('''ALTER TABLE core_device."DeviceInfo"
                  ALTER COLUMN "lastOnline" TYPE TIMESTAMP WITH TIME ZONE
                  USING "lastOnline" AT TIME ZONE 'UTC'
               ''')

    op.execute('''ALTER TABLE core_device."DeviceInfo"
                  ALTER COLUMN "lastUpdateCheck" TYPE TIMESTAMP WITH TIME ZONE
                  USING "lastUpdateCheck" AT TIME ZONE 'UTC'
               ''')

    op.execute('''ALTER TABLE core_device."DeviceInfo"
                  ALTER COLUMN "createdDate" TYPE TIMESTAMP WITH TIME ZONE
                  USING "createdDate" AT TIME ZONE 'UTC'
               ''')


    # core_device."DeviceUpdate"
    op.execute('''ALTER TABLE core_device."DeviceUpdate"
                  ALTER COLUMN "buildDate" TYPE TIMESTAMP WITH TIME ZONE
                  USING "buildDate" AT TIME ZONE 'UTC'
               ''')

def downgrade():
    pass


import logging
from datetime import datetime
from typing import Union

import pytz
from twisted.internet.defer import Deferred

from peek_core_device._private.storage.DeviceInfoTuple import DeviceInfoTuple
from peek_core_device._private.storage.DeviceUpdateTuple import DeviceUpdateTuple
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)

class DeviceUpdateTupleProvider(TuplesProviderABC):
    def __init__(self, ormSessionCreator):
        self._ormSessionCreator = ormSessionCreator

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict,
                      tupleSelector: TupleSelector) -> Union[Deferred, bytes]:
        # Potential filters can be placed here.
        deviceId = tupleSelector.selector.get("deviceId")

        ormSession = self._ormSessionCreator()
        try:

            deviceInfo = None
            if deviceId is not None:
                deviceInfo = (
                    ormSession.query(DeviceInfoTuple)
                        .filter(DeviceInfoTuple.deviceId == deviceId)
                        .one()
                )

                deviceInfo.lastUpdateCheck = datetime.now(pytz.utc)
                ormSession.commit()

            tuples = []

            # Return just the one, latest update, if this is a device checking
            if deviceInfo:
                updates = (
                    ormSession.query(DeviceUpdateTuple)
                        .filter(DeviceUpdateTuple.deviceType == deviceInfo.deviceType)
                        .filter(DeviceUpdateTuple.isEnabled == True)
                        .order_by(DeviceUpdateTuple.buildDate)
                        .all()
                )

                if updates:
                    update = updates[-1]
                    if update.updateVersion != deviceInfo.updateVersion:
                        tuples = [update]

            # Else, this is the admin interface, return all of them.
            else:
                tuples = (
                    ormSession.query(DeviceUpdateTuple)
                        .order_by(DeviceUpdateTuple.buildDate)
                        .all()
                )

            # Create the vortex message
            return Payload(filt, tuples=tuples).makePayloadEnvelope().toVortexMsg()

        finally:
            ormSession.close()



import logging
from typing import List

from sqlalchemy.orm.exc import NoResultFound
from twisted.internet.defer import Deferred

from peek_core_device._private.server.controller.NotifierController import \
    NotifierController
from peek_core_device._private.storage.DeviceInfoTuple import DeviceInfoTuple
from peek_core_device._private.tuples.UpdateDeviceOnlineAction import \
    UpdateDeviceOnlineAction
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Tuple import Tuple
from vortex.TupleAction import TupleActionABC

logger = logging.getLogger(__name__)


class OnlineController:
    def __init__(self, dbSessionCreator, notifierController: NotifierController):
        self._dbSessionCreator = dbSessionCreator
        self._notifierController = notifierController

        self._setAllDevicesOfflineBlocking()

    def shutdown(self):
        pass

    def processTupleAction(self, tupleAction: TupleActionABC) -> List[Tuple]:

        if isinstance(tupleAction, UpdateDeviceOnlineAction):
            return self._processUpdateOnline(tupleAction)

    @deferToThreadWrapWithLogger(logger)
    def _processUpdateOnline(self, action: UpdateDeviceOnlineAction) -> List[Tuple]:
        """ Process Online Status Update

        :rtype: Deferred
        """
        session = self._dbSessionCreator()
        try:
            deviceInfo = (
                session.query(DeviceInfoTuple)
                    .filter(DeviceInfoTuple.deviceId == action.deviceId)
                    .one()
            )

            deviceId = deviceInfo.deviceId

            deviceInfo.lastOnline = action.dateTime
            deviceInfo.isOnline = action.isOnline

            session.commit()

            self._notifierController.notifyDeviceInfo(deviceId=deviceId)
            self._notifierController.notifyDeviceOnline(
                deviceInfo.deviceId,
                deviceInfo.deviceToken,
                deviceInfo.isOnline
            )

            return []

        except NoResultFound:
            return []

        finally:
            session.close()

    def _setAllDevicesOfflineBlocking(self):
        """ Set All Devices to Offline

        """
        session = self._dbSessionCreator()
        try:
            session.execute(DeviceInfoTuple.__table__.update().values(isOnline=False))
            session.commit()

        finally:
            session.close()



from peek_core_device._private.PluginNames import deviceTuplePrefix
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship
from sqlalchemy.orm.collections import attribute_mapped_collection
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.sql.expression import cast, null, case
from sqlalchemy.sql.schema import Index
from sqlalchemy.types import Integer, String, Boolean
from vortex.Tuple import Tuple, addTupleType

from .DeclarativeBase import DeclarativeBase

"""Mapping a polymorphic-valued vertical table as a dictionary.

For any given properties row, the value of the 'type' column will point to the
'_value' column active for that row.

It only differs in the mapping for vertical rows.  Here,
we'll use a @hybrid_property to attune a smart '.value' attribute that wraps up
reading and writing those various '_value' columns and keeps the '.type' up to
date.

"""

from sqlalchemy.orm.interfaces import PropComparator
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import event
from sqlalchemy import literal_column


class ProxiedDictMixin(object):
    """Adds obj[key] access to a mapped class.

    This class basically proxies dictionary access to an attribute
    called ``_proxied``.  The class which inherits this class
    should have an attribute called ``_proxied`` which points to a dictionary.

    """
    _proxied = None

    def __len__(self):
        return len(self._proxied)

    def __iter__(self):
        return iter(self._proxied)

    def __getitem__(self, key):
        return self._proxied[str(key)]

    def __contains__(self, key):
        return str(key) in self._proxied

    def __setitem__(self, key, value):
        self._proxied[str(key)] = value

    def __delitem__(self, key):
        del self._proxied[str(key)]


class PolymorphicVerticalProperty(object):
    """A key/value pair with polymorphic value storage.

    The class which is mapped should indicate typing information
    within the "info" dictionary of mapped Column objects; see
    the AnimalFact mapping below for an example.

    """

    def __init__(self, key, value=None):
        self.key = key
        self.value = value

    @hybrid_property
    def value(self):
        fieldname, discriminator = self.type_map[self.type]
        if fieldname is None:
            return None
        else:
            return getattr(self, fieldname)

    @value.setter
    def value(self, value):
        py_type = type(value)
        fieldname, discriminator = self.type_map[py_type]

        self.type = discriminator
        if fieldname is not None:
            setattr(self, fieldname, value)

    @value.deleter
    def value(self):
        self._set_value(None)

    @value.comparator
    class value(PropComparator):
        """A comparator for .value, builds a polymorphic comparison via CASE.

        """

        def __init__(self, cls):
            self.cls = cls

        def _case(self):
            pairs = set(self.cls.type_map.values())
            whens = [
                (
                    literal_column("'%s'" % discriminator),
                    cast(getattr(self.cls, attribute), String)
                ) for attribute, discriminator in pairs
                if attribute is not None
            ]
            return case(whens, self.cls.type, null())

        def __eq__(self, other):
            return self._case() == cast(other, String)

        def __ne__(self, other):
            return self._case() != cast(other, String)

    def __repr__(self):
        return '<%s %r=%r>' % (self.__class__.__name__, self.key, self.value)


@addTupleType
class SettingProperty(PolymorphicVerticalProperty, Tuple, DeclarativeBase):
    """A setting property."""

    __tablename__ = 'SettingProperty'
    __tupleType__ = deviceTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    settingId = Column(ForeignKey('Setting.id'), nullable=False)
    key = Column(String(50), nullable=False)
    type = Column(String(16))

    # add information about storage for different types
    # in the info dictionary of Columns
    int_value = Column(Integer, info={'type': (int, 'integer')})
    char_value = Column(String, info={'type': (str, 'string')})
    boolean_value = Column(Boolean, info={'type': (bool, 'boolean')})

    def __init__(self, key=None, value=None):
        PolymorphicVerticalProperty.__init__(self, key=key, value=value)
        Tuple.__init__(self)

    __table_args__ = (
        Index("idx_SettingProperty_settingId", settingId),
    )


@addTupleType
class Setting(ProxiedDictMixin, Tuple, DeclarativeBase):
    """an Animal"""

    __tablename__ = 'Setting'
    __tupleType__ = deviceTuplePrefix + __tablename__

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String)

    properties = relationship("SettingProperty",
                              collection_class=attribute_mapped_collection('key'))

    propertyObjects = relationship("SettingProperty", viewonly=True, lazy=False)

    _proxied = association_proxy("properties", "value",
                                 creator=
                                 lambda key, value: SettingProperty(key=key,
                                                                    value=value))

    def __init__(self, name=None):
        self.name = name

    def __repr__(self):
        return "Setting(%r)" % self.name

    @classmethod
    def with_characteristic(self, key, value):
        return self.properties.any(key=key, value=value)


@event.listens_for(SettingProperty, "mapper_configured",
                   propagate=True)
def on_new_class(mapper, cls_):
    """Look for Column objects with type info in them, and work up
    a lookup table."""

    info_dict = {}
    info_dict[type(None)] = (None, 'none')
    info_dict['none'] = (None, 'none')

    for k in list(mapper.c.keys()):
        col = mapper.c[k]
        if 'type' in col.info:
            python_type, discriminator = col.info['type']
            info_dict[python_type] = (k, discriminator)
            info_dict[discriminator] = (k, discriminator)
    cls_.type_map = info_dict


class PropertyKey(object):
    def __init__(self, name, defaultValue, propertyDict):
        self.name = name
        self.defaultValue = defaultValue
        propertyDict[name] = self

    def __repr__(self):
        return self.name


def _getSetting(ormSession, name, propertyDict, key=None, value=None):
    all = ormSession.query(Setting).filter(Setting.name == name).all()

    if all:
        setting = all[0]
        ormSession.expire(setting)
    else:
        setting = Setting(name)
        ormSession.add(setting)
        ormSession.commit()

    needsCommit = False

    for prop in list(propertyDict.values()):
        if not prop.name in setting:
            setting[prop.name] = prop.defaultValue
            needsCommit = True

    if needsCommit:
        ormSession.commit()

    if not key:
        return setting

    # Make sure the propery is defined for this setting
    assert str(key) in propertyDict, "Key %s is not defined in setting %s" % (key, name)

    if value is None:
        return setting[key]

    setting[key] = value
    ormSession.commit()

    # Close after return
    return setting[key]


# =============================================================================
# GLOBAL PROPERTIES
# =============================================================================

globalProperties = {}


def globalSetting(ormSession, key=None, value=None):
    return _getSetting(ormSession, "Global", globalProperties, key=key, value=value)


AUTO_ENROLLMENT = PropertyKey('Device Auto Enrollment', False,
                              propertyDict=globalProperties)

# =============================================================================
# CLIENT PROPERTIES
# =============================================================================


FIELD_ENROLLMENT_ENABLED = PropertyKey('Field Enrollment Enabled', True,
                                       propertyDict=globalProperties)

OFFICE_ENROLLMENT_ENABLED = PropertyKey('Office Enrollment Enabled', False,
                                        propertyDict=globalProperties)
